@@ -1,3 +1,4 @@
+//go:build linux
 // +build linux
 
 package server
@@ -13,7 +14,6 @@ import (
 func TestAddOCIBindsForDev(t *testing.T) {
 	specgen, err := generate.New("linux")
 	if err != nil {
-
 		t.Error(err)
 	}
 	config := &pb.ContainerConfig{
@@ -24,7 +24,7 @@ func TestAddOCIBindsForDev(t *testing.T) {
 			},
 		},
 	}
-	_, binds, err := addOCIBindMounts(context.Background(), "", config, &specgen, "")
+	_, binds, err := addOCIBindMounts(context.Background(), "", config, &specgen, "", false)
 	if err != nil {
 		t.Error(err)
 	}
@@ -48,7 +48,6 @@ func TestAddOCIBindsForDev(t *testing.T) {
 func TestAddOCIBindsForSys(t *testing.T) {
 	specgen, err := generate.New("linux")
 	if err != nil {
-
 		t.Error(err)
 	}
 	config := &pb.ContainerConfig{
@@ -59,7 +58,7 @@ func TestAddOCIBindsForSys(t *testing.T) {
 			},
 		},
 	}
-	_, binds, err := addOCIBindMounts(context.Background(), "", config, &specgen, "")
+	_, binds, err := addOCIBindMounts(context.Background(), "", config, &specgen, "", false)
 	if err != nil {
 		t.Error(err)
 	}
