@@ -24,6 +24,7 @@ import (
 	libsandbox "github.com/cri-o/cri-o/internal/lib/sandbox"
 	"github.com/cri-o/cri-o/internal/log"
 	oci "github.com/cri-o/cri-o/internal/oci"
+	"github.com/cri-o/cri-o/internal/resourcestore"
 	ann "github.com/cri-o/cri-o/pkg/annotations"
 	libconfig "github.com/cri-o/cri-o/pkg/config"
 	"github.com/cri-o/cri-o/pkg/sandbox"
@@ -308,14 +309,14 @@ func (s *Server) runPodSandbox(ctx context.Context, req *pb.RunPodSandboxRequest
 		return nil, errors.Wrap(err, "setting pod sandbox name and id")
 	}
 
-	cleanupFuncs := make([]func(), 0)
+	resourceCleaner := resourcestore.NewResourceCleaner()
 	defer func() {
 		// no error, no need to cleanup
 		if retErr == nil || isContextError(retErr) {
 			return
 		}
-		for i := len(cleanupFuncs) - 1; i >= 0; i-- {
-			cleanupFuncs[i]()
+		if err := resourceCleaner.Cleanup(); err != nil {
+			log.Errorf(ctx, "Unable to cleanup: %v", err)
 		}
 	}()
 
@@ -327,9 +328,11 @@ func (s *Server) runPodSandbox(ctx context.Context, req *pb.RunPodSandboxRequest
 		return nil, errors.Wrapf(err, resourceErr.Error())
 	}
 
-	cleanupFuncs = append(cleanupFuncs, func() {
-		log.Infof(ctx, "runSandbox: releasing pod sandbox name: %s", sbox.Name())
+	description := fmt.Sprintf("runSandbox: releasing pod sandbox name: %s", sbox.Name())
+	resourceCleaner.Add(ctx, description, func() error {
+		log.Infof(ctx, description)
 		s.ReleasePodName(sbox.Name())
+		return nil
 	})
 
 	kubeAnnotations := sbox.Config().GetAnnotations()
@@ -351,9 +354,11 @@ func (s *Server) runPodSandbox(ctx context.Context, req *pb.RunPodSandboxRequest
 	if err != nil {
 		return nil, err
 	}
-	cleanupFuncs = append(cleanupFuncs, func() {
-		log.Infof(ctx, "runSandbox: releasing container name: %s", containerName)
+	description = fmt.Sprintf("runSandbox: releasing container name: %s", containerName)
+	resourceCleaner.Add(ctx, description, func() error {
+		log.Infof(ctx, description)
 		s.ReleaseContainerName(containerName)
+		return nil
 	})
 
 	var labelOptions []string
@@ -389,11 +394,14 @@ func (s *Server) runPodSandbox(ctx context.Context, req *pb.RunPodSandboxRequest
 	if err != nil {
 		return nil, fmt.Errorf("error creating pod sandbox with name %q: %v", sbox.Name(), err)
 	}
-	cleanupFuncs = append(cleanupFuncs, func() {
-		log.Infof(ctx, "runSandbox: removing pod sandbox from storage: %s", sbox.ID())
-		if err2 := s.StorageRuntimeServer().RemovePodSandbox(sbox.ID()); err2 != nil {
+	description = fmt.Sprintf("runSandbox: removing pod sandbox from storage: %s", sbox.ID())
+	resourceCleaner.Add(ctx, description, func() error {
+		log.Infof(ctx, description)
+		err2 := s.StorageRuntimeServer().DeleteContainer(sbox.ID())
+		if err2 != nil {
 			log.Warnf(ctx, "could not cleanup pod sandbox %q: %v", sbox.ID(), err2)
 		}
+		return err2
 	})
 
 	// set log directory
@@ -548,11 +556,14 @@ func (s *Server) runPodSandbox(ctx context.Context, req *pb.RunPodSandboxRequest
 			return nil, err
 		}
 		pathsToChown = append(pathsToChown, shmPath)
-		cleanupFuncs = append(cleanupFuncs, func() {
-			log.Infof(ctx, "runSandbox: unmounting shmPath for sandbox %s", sbox.ID())
-			if err2 := unix.Unmount(shmPath, unix.MNT_DETACH); err2 != nil {
+		description = fmt.Sprintf("runSandbox: unmounting shmPath for sandbox %s", sbox.ID())
+		resourceCleaner.Add(ctx, description, func() error {
+			log.Infof(ctx, description)
+			err2 := unix.Unmount(shmPath, unix.MNT_DETACH)
+			if err2 != nil {
 				log.Warnf(ctx, "failed to unmount shm for sandbox: %v", err2)
 			}
+			return err2
 		})
 	}
 
@@ -574,11 +585,18 @@ func (s *Server) runPodSandbox(ctx context.Context, req *pb.RunPodSandboxRequest
 		return nil, err
 	}
 
-	cleanupFuncs = append(cleanupFuncs, func() {
-		log.Infof(ctx, "runSandbox: deleting container ID from idIndex for sandbox %s", sbox.ID())
-		if err2 := s.CtrIDIndex().Delete(sbox.ID()); err2 != nil {
-			log.Warnf(ctx, "could not delete ctr id %s from idIndex", sbox.ID())
+	description = fmt.Sprintf("runSandbox: deleting container ID from idIndex for sandbox %s", sbox.ID())
+	resourceCleaner.Add(ctx, description, func() error {
+		log.Infof(ctx, description)
+		err2 := s.CtrIDIndex().Delete(sbox.ID())
+		if err2 != nil {
+			// already deleted
+			if strings.Contains(err2.Error(), noSuchID) {
+				return nil
+			}
+			log.Warnf(ctx, "Could not delete ctr id %s from idIndex: %v", sbox.ID(), err2)
 		}
+		return err2
 	})
 
 	// set log path inside log directory
@@ -671,22 +689,31 @@ func (s *Server) runPodSandbox(ctx context.Context, req *pb.RunPodSandboxRequest
 	if err := s.addSandbox(sb); err != nil {
 		return nil, err
 	}
-	cleanupFuncs = append(cleanupFuncs, func() {
-		log.Infof(ctx, "runSandbox: removing pod sandbox %s", sbox.ID())
-		if err := s.removeSandbox(sbox.ID()); err != nil {
+	description = fmt.Sprintf("runSandbox: removing pod sandbox %s", sbox.ID())
+	resourceCleaner.Add(ctx, description, func() error {
+		log.Infof(ctx, description)
+		err := s.removeSandbox(sbox.ID())
+		if err != nil {
 			log.Warnf(ctx, "could not remove pod sandbox: %v", err)
 		}
+		return err
 	})
 
 	if err := s.PodIDIndex().Add(sbox.ID()); err != nil {
 		return nil, err
 	}
 
-	cleanupFuncs = append(cleanupFuncs, func() {
-		log.Infof(ctx, "runSandbox: deleting pod ID %s from idIndex", sbox.ID())
-		if err := s.PodIDIndex().Delete(sbox.ID()); err != nil {
-			log.Warnf(ctx, "could not delete pod id %s from idIndex", sbox.ID())
+	description = fmt.Sprintf("runSandbox: deleting pod ID %s from idIndex", sbox.ID())
+	resourceCleaner.Add(ctx, description, func() error {
+		log.Infof(ctx, description)
+		err := s.PodIDIndex().Delete(sbox.ID())
+		if err != nil {
+			if strings.Contains(err.Error(), noSuchID) {
+				return nil
+			}
+			log.Warnf(ctx, "Could not delete pod id %s from idIndex: %v", sbox.ID(), err)
 		}
+		return err
 	})
 
 	for k, v := range kubeAnnotations {
@@ -702,15 +729,19 @@ func (s *Server) runPodSandbox(ctx context.Context, req *pb.RunPodSandboxRequest
 	// set up namespaces
 	nsCleanupFuncs, err := s.configureGeneratorForSandboxNamespaces(hostNetwork, hostIPC, hostPID, sandboxIDMappings, sysctls, sb, g)
 	// We want to cleanup after ourselves if we are managing any namespaces and fail in this function.
-	cleanupFuncs = append(cleanupFuncs, func() {
-		log.Infof(ctx, "runSandbox: cleaning up namespaces after failing to run sandbox %s", sbox.ID())
+	nsCleanupDescription := fmt.Sprintf("runSandbox: cleaning up namespaces after failing to run sandbox %s", sbox.ID())
+	nsCleanupFunc := func() error {
+		log.Infof(ctx, description)
 		for idx := range nsCleanupFuncs {
 			if err2 := nsCleanupFuncs[idx](); err2 != nil {
 				log.Infof(ctx, "runSandbox: failed to cleanup namespace: %s", err2.Error())
+				return err2
 			}
 		}
-	})
+		return nil
+	}
 	if err != nil {
+		resourceCleaner.Add(ctx, nsCleanupDescription, nsCleanupFunc)
 		return nil, err
 	}
 
@@ -721,13 +752,21 @@ func (s *Server) runPodSandbox(ctx context.Context, req *pb.RunPodSandboxRequest
 	if s.config.ManageNSLifecycle {
 		ips, result, err = s.networkStart(ctx, sb)
 		if err != nil {
+			resourceCleaner.Add(ctx, nsCleanupDescription, nsCleanupFunc)
 			return nil, err
 		}
-		cleanupFuncs = append(cleanupFuncs, func() {
-			log.Infof(ctx, "runSandbox: in manageNSLifecycle, stopping network for sandbox %s", sb.ID())
-			if err2 := s.networkStop(ctx, sb); err2 != nil {
-				log.Errorf(ctx, "error stopping network on cleanup: %v", err2)
+		description = fmt.Sprintf("runSandbox: stopping network for sandbox %s", sb.ID())
+		resourceCleaner.Add(ctx, description, func() error {
+			log.Infof(ctx, description)
+			// use a new context to prevent an expired context from preventing a stop
+			if err := s.networkStop(context.Background(), sb); err != nil {
+				log.Errorf(ctx, "error stopping network on cleanup: %v", err)
+				return err
 			}
+
+			// Now that we've succeeded in stopping the network, cleanup namespaces
+			log.Infof(ctx, nsCleanupDescription)
+			return nsCleanupFunc()
 		})
 		if result != nil {
 			resultCurrent, err := current.NewResultFromResult(result)
@@ -759,11 +798,14 @@ func (s *Server) runPodSandbox(ctx context.Context, req *pb.RunPodSandboxRequest
 	if err != nil {
 		return nil, fmt.Errorf("failed to mount container %s in pod sandbox %s(%s): %v", containerName, sb.Name(), sbox.ID(), err)
 	}
-	cleanupFuncs = append(cleanupFuncs, func() {
-		log.Infof(ctx, "runSandbox: stopping storage container for sandbox %s", sbox.ID())
-		if err2 := s.StorageRuntimeServer().StopContainer(sbox.ID()); err2 != nil {
+	description = fmt.Sprintf("runSandbox: stopping storage container for sandbox %s", sbox.ID())
+	resourceCleaner.Add(ctx, description, func() error {
+		log.Infof(ctx, description)
+		err2 := s.StorageRuntimeServer().StopContainer(sbox.ID())
+		if err2 != nil {
 			log.Warnf(ctx, "could not stop storage container: %v: %v", sbox.ID(), err2)
 		}
+		return err2
 	})
 	g.AddAnnotation(annotations.MountPoint, mountPoint)
 
@@ -900,9 +942,11 @@ func (s *Server) runPodSandbox(ctx context.Context, req *pb.RunPodSandboxRequest
 	}
 
 	s.addInfraContainer(container)
-	cleanupFuncs = append(cleanupFuncs, func() {
-		log.Infof(ctx, "runSandbox: removing infra container %s", container.ID())
+	description = fmt.Sprintf("runSandbox: removing infra container %s", container.ID())
+	resourceCleaner.Add(ctx, description, func() error {
+		log.Infof(ctx, description)
 		s.removeInfraContainer(container)
+		return nil
 	})
 
 	if sandboxIDMappings != nil {
@@ -925,23 +969,25 @@ func (s *Server) runPodSandbox(ctx context.Context, req *pb.RunPodSandboxRequest
 		return nil, err
 	}
 
-	cleanupFuncs = append(cleanupFuncs, func() {
+	description = fmt.Sprintf("runSandbox: stopping container %s", container.ID())
+	resourceCleaner.Add(ctx, description, func() error {
 		// Clean-up steps from RemovePodSanbox
-		log.Infof(ctx, "runSandbox: stopping container %s", container.ID())
-		if err2 := s.Runtime().StopContainer(ctx, container, int64(10)); err2 != nil {
-			log.Warnf(ctx, "failed to stop container %s: %v", container.Name(), err2)
-		}
-		if err2 := s.Runtime().WaitContainerStateStopped(ctx, container); err2 != nil {
-			log.Warnf(ctx, "failed to get container 'stopped' status %s in pod sandbox %s: %v", container.Name(), sb.ID(), err2)
+		log.Infof(ctx, description)
+		if err := s.ContainerServer.StopContainer(ctx, container, int64(10)); err != nil {
+			return errors.Errorf("failed to stop container for removal")
 		}
+
 		log.Infof(ctx, "runSandbox: deleting container %s", container.ID())
 		if err2 := s.Runtime().DeleteContainer(container); err2 != nil {
 			log.Warnf(ctx, "failed to delete container %s in pod sandbox %s: %v", container.Name(), sb.ID(), err2)
+			return err2
 		}
 		log.Infof(ctx, "runSandbox: writing container %s state to disk", container.ID())
 		if err2 := s.ContainerStateToDisk(container); err2 != nil {
 			log.Warnf(ctx, "failed to write container state %s in pod sandbox %s: %v", container.Name(), sb.ID(), err2)
+			return err2
 		}
+		return nil
 	})
 
 	if err := s.ContainerStateToDisk(container); err != nil {
@@ -953,11 +999,14 @@ func (s *Server) runPodSandbox(ctx context.Context, req *pb.RunPodSandboxRequest
 		if err != nil {
 			return nil, err
 		}
-		cleanupFuncs = append(cleanupFuncs, func() {
-			log.Infof(ctx, "runSandbox: stopping network for sandbox %s when not manageNSLifecycle", sb.ID())
-			if err2 := s.networkStop(ctx, sb); err2 != nil {
+		description = fmt.Sprintf("runSandbox: stopping network for sandbox %s when not manage NS", sb.ID())
+		resourceCleaner.Add(ctx, description, func() error {
+			log.Infof(ctx, description)
+			err2 := s.networkStop(context.Background(), sb)
+			if err2 != nil {
 				log.Errorf(ctx, "error stopping network on cleanup: %v", err2)
 			}
+			return err2
 		})
 	}
 
@@ -967,7 +1016,7 @@ func (s *Server) runPodSandbox(ctx context.Context, req *pb.RunPodSandboxRequest
 	sb.AddIPs(ips)
 
 	if isContextError(ctx.Err()) {
-		if err := s.resourceStore.Put(sbox.Name(), sb, cleanupFuncs); err != nil {
+		if err := s.resourceStore.Put(sbox.Name(), sb, resourceCleaner); err != nil {
 			log.Errorf(ctx, "runSandbox: failed to save progress of sandbox %s: %v", sbox.ID(), err)
 		}
 		log.Infof(ctx, "runSandbox: context was either canceled or the deadline was exceeded: %v", ctx.Err())
@@ -1031,7 +1080,7 @@ func (s *Server) configureGeneratorForSysctls(ctx context.Context, g generate.Ge
 
 	for _, sysctl := range defaultSysctls {
 		if err := sysctl.Validate(hostNetwork, hostIPC); err != nil {
-			log.Warnf(ctx, "skipping invalid sysctl %s: %v", sysctl, err)
+			log.Warnf(ctx, "Skipping invalid sysctl specified by config %s: %v", sysctl, err)
 			continue
 		}
 		g.AddLinuxSysctl(sysctl.Key(), sysctl.Value())
@@ -1041,6 +1090,11 @@ func (s *Server) configureGeneratorForSysctls(ctx context.Context, g generate.Ge
 	// extract linux sysctls from annotations and pass down to oci runtime
 	// Will override any duplicate default systcl from crio.conf
 	for key, value := range sysctls {
+		sysctl := libconfig.NewSysctl(key, value)
+		if err := sysctl.Validate(hostNetwork, hostIPC); err != nil {
+			log.Warnf(ctx, "Skipping invalid sysctl specified over CRI %s: %v", sysctl, err)
+			continue
+		}
 		g.AddLinuxSysctl(key, value)
 		sysctlsToReturn[key] = value
 	}
